@@ -147,6 +147,9 @@ class Normalise:
         # Which rows to np.nan in one go
         nan_array = []
 
+        # Which rows to none in one go
+        none_array = []
+
         error_file = None
 
         # regex to match the string to_file in between quotes or double quotes
@@ -168,10 +171,14 @@ class Normalise:
             if "action" in rules[j]:
                 action = rules[j]["action"]
 
-            if action == "np.nan":
+            if action in ("np.nan", "np_nan", "nan", "NaN", "np-nan"):
                 # Which rows to np.nan later on in one go
                 nan_array.append(i)
 
+            if action in ("None", "none"):
+                # Which rows to None later on in one go
+                none_array.append(i)
+
             if action == "drop":
                 # Which rows to drop later onin one go
                 drop_array.append(i)
@@ -200,23 +207,65 @@ class Normalise:
                 if "check_coverage" in rules[j]:
                     global_check_coverage = rules[j]["check_coverage"]
 
-                if ("reset_coverage" in rules[j]):
+                if "reset_coverage" in rules[j]:
                     if (bool)(rules[j]["reset_coverage"]):
                         reset_coverage = bool(rules[j]["reset_coverage"])
 
-                if ("action" in rules[j]):
+                if "action" in rules[j]:
                     global_action = rules[j]["action"]
 
-                if ("verbose" in rules[j]):
+                if "verbose" in rules[j]:
                     global_verbose = rules[j]["verbose"]
 
-                if ("column_mapping" in rules[j]):
-                    dataframe = dataframe.rename(
-                        columns=rules[j]["column_mapping"])
+                if "column_mapping" in rules[j]:
+                    for old, new in rules[j]["column_mapping"].items():
+                        dataframe[new] = dataframe[old]
 
-                if ("mapping" in rules[j]):
+                if "mapping" in rules[j]:
                     global_mapping = rules[j]["mapping"]
 
+                if "drop_duplicates" in rules[j]:
+                    dataframe.drop_duplicates(
+                        subset=rules[j]["drop_duplicates"], inplace=True)
+
+                if "shared_colname_drop" in rules[j]:
+                    dups = dataframe.T.duplicated()
+                    true_values = dups.loc[dups]
+                    print("Dropping: " + str(true_values.loc[dups].keys().tolist()))
+
+                    dataframe = dataframe.drop(dataframe.loc[:, dups], axis=1)
+
+                if "concat" in rules[j]:
+                    name = rules[j]["concat"]["name"]
+                    dataframe[name] = ""
+                    for column in rules[j]["concat"]["columns"]:
+                        dataframe[name] += dataframe[column].astype(str)
+
+                if "fillna" in rules[j]:
+                    dataframe = dataframe.fillna(rules[j]["fillna"])
+                
+                if "fillna_median" in rules[j]:
+                    dataframe = dataframe.fillna(dataframe.median())
+
+                if "fillna_mean" in rules[j]:
+                    dataframe = dataframe.fillna(dataframe.mean())
+                
+                if "fillna_diffcol" in rules[j]:
+                    dataframe = dataframe.fillna(dataframe[rules[j]["fillna_diffcol"]])
+
+                if "operator" in rules[j]:
+                    operator = rules[j]["operator"]["type"]
+                    column = rules[j]["operator"]["columns"]
+                    name = rules[j]["operator"]["name"]
+                    try: 
+                        if operator == "divide":
+                            dataframe[name] = dataframe[column[0]]/dataframe[column[1]]
+
+                        if operator == "multiply":
+                            dataframe[name] = dataframe[column[0]] * dataframe[column[1]]
+                    except TypeError:
+                        print("Operator failed: expected numbers")
+
                 # Continue with next rule
                 return True
 
@@ -288,9 +337,27 @@ class Normalise:
                             handle_mismatch(i, value, column_values)
                     except ValueError:
                         handle_mismatch(i, value, column_values)
+            
+            if "timestamp" in rules[j]:
+                try:
+                    dataframe[rules[j]["column"]] = pd.to_datetime(dataframe[rules[j]["column"]], utc=True).dt.strftime(rules[j]["timestamp"])
+                except TypeError:
+                    print("Timestamping failed: " + str(rules[j]["column"]))
+
+            if "range_time" in rules[j]:
+                left = pd.to_datetime(rules[j]["range_time"][0], utc=True)
+                right = pd.to_datetime(rules[j]["range_time"][1], utc=True)
+
+                for i, value in column_values.items():
+                    try:
+                        value = pd.to_datetime(value, utc=True)
+                        if not (left <= value <= right):
+                            handle_mismatch(i, value, column_values)
+                    except ValueError:
+                        handle_mismatch(i, value, column_values)
 
             if "type" in rules[j]:
-
+                    
                 if rules[j]["type"] == "percentage":
                     for i, value in column_values.items():
                         # Regex to match float between 0 and 100
@@ -390,9 +457,32 @@ class Normalise:
             # Make all the indexes inside the np.nan array nan values
             column_values.loc[nan_array] = np.nan
 
+            # Make all the indexes inside the none array None values
+            column_values.loc[none_array] = None
+
+            # If there are any values in drop_array, nan_aray or none_array print the column name and the length of the array
+            for array in (drop_array, none_array, nan_array):
+                if len(array) > 0:
+                    array_name = 'dropped: ' if array == drop_array else 'made None: ' if array == none_array else 'made NaN: '
+                    print(f"Column {rules[j]['column']}, {array_name}{len(array)} values")
+
+            try:
+                if "fillna" in rules[j]:
+                    column_values = column_values.fillna(rules[j]["fillna"])
+                
+                if "fillna_median" in rules[j]:
+                    column_values = column_values.fillna(column_values.median())
+
+                if "fillna_mean" in rules[j]:
+                    column_values = column_values.fillna(column_values.mean())
+                
+                if "fillna_diffcol" in rules[j]:
+                    column_values = column_values.fillna(dataframe[rules[j]["fillna_diffcol"]])
+            except TypeError:
+                pass
+
             # Update final dataframe with cached values
             dataframe[rules[j]["column"]].update(column_values)
-    
 
             errors_found = False
         return dataframe
